from __future__ import annotations

import re
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from itertools import filterfalse
from types import MappingProxyType
from typing import Protocol, runtime_checkable

from typing_extensions import Self, TypeAlias

from avilla.core.platform import Land

EMPTY_MAP = MappingProxyType({})
ESCAPE = {")": "\\)", "(": "\\(", "]": "\\]", "[": "\\[", "}": "\\}", "{": "\\{"}

_follows_pattern = re.compile(r"(?P<name>(\w+?|[*~]))(#(?P<predicate>\w+))?(\((?P<literal>[^#]+?)\))?")
FollowsPredicater: TypeAlias = "Callable[[str], bool]"


@dataclass
class _FollowItem:
    name: str
    literal: str | None = None
    predicate: FollowsPredicater | None = None


def _parse_follows_item(item: str, items: dict[str, _FollowItem], predicates: dict[str, FollowsPredicater]):
    if item.startswith("::"):
        if "land" in items:
            raise ValueError("land already exists")
        item = item[2:]
        items["land"] = _FollowItem("land")
        if not item.strip():
            return
    if "*" in items:
        raise ValueError("wildcard already exists, no more items allowed")
    if not (m := _follows_pattern.fullmatch(item)):
        raise ValueError(f"invalid item: {item}")
    name = m["name"]
    if name == "~" and items:
        raise ValueError("inherit can only be the first item and only one")
    if m["literal"] and m["predicate"]:
        raise ValueError(f"duplicate literal and predicate: {item}")
    if m["literal"] is not None:
        literal = m["literal"].translate(str.maketrans(ESCAPE))
    else:
        literal = None
    predicate = predicates.get(m["predicate"]) if m["predicate"] else None
    items[name] = _FollowItem(name, literal, predicate)


def _parse_follows(pattern: str, **kwargs: FollowsPredicater) -> list[_FollowItem]:
    items = {}
    item = ""
    bracket_stack = []
    for i, char in enumerate(pattern):
        if char == "." and not bracket_stack:
            _parse_follows_item(item, items, kwargs)
            item = ""
            continue
        if char == "(":
            bracket_stack.append(i)
        elif char == ")":
            if not bracket_stack:
                raise ValueError(f"Unclosed parenthesis: {item})")
            bracket_stack.pop()
        item += char
    if bracket_stack:
        raise ValueError(f"Unclosed parenthesis: {pattern[bracket_stack[0]:]}")
    if item:
        _parse_follows_item(item, items, kwargs)
    return list(items.values())


class Selector:
    pattern: Mapping[str, str]

    def __init__(self, pattern: Mapping[str, str] = EMPTY_MAP) -> None:
        self.pattern = MappingProxyType({**pattern})

    def __getattr__(self, name: str) -> Callable[[str], Self]:
        def wrapper(content: str) -> Self:
            return Selector(pattern={**self.pattern, name: content})

        return wrapper

    def __hash__(self) -> int:
        return hash(("Selector", *self.pattern.items()))

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Selector) and o.pattern == self.pattern

    def __contains__(self, key: str) -> bool:
        return key in self.pattern

    def __getitem__(self, key: str) -> str:
        return self.pattern[key]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}().{'.'.join(f'{k}({v})' for k, v in self.pattern.items())}"

    @property
    def empty(self) -> bool:
        return not self.pattern

    @property
    def path(self) -> str:
        return ".".join(self.pattern)

    @property
    def path_without_land(self) -> str:
        return ".".join(filterfalse(lambda x: x == "land", self.pattern))

    @property
    def last_key(self) -> str:
        return next(reversed(self.pattern.keys()))

    @property
    def last_value(self) -> str:
        return next(reversed(self.pattern.values()))

    def items(self):
        return self.pattern.items()

    def appendix(self, key: str, value: str):
        return Selector(pattern={**self.pattern, key: value})

    def land(self, land: Land | str):
        if isinstance(land, Land):
            land = land.name

        return Selector({"land": land, **{k: v for k, v in self.pattern.items() if k != "land"}})

    def to_selector(self):
        return self

    @classmethod
    def from_follows_pattern(cls, pattern: str):
        items = _parse_follows(pattern)
        mapping = {}
        for i in items:
            if i.literal is None:
                raise ValueError("literal expected")
            mapping[i.name] = i.literal
        return cls(mapping)

    def follows(self, pattern: str, **kwargs: FollowsPredicater) -> bool:
        items = _parse_follows(pattern, **kwargs)
        index = 0
        for index, (item, name, value) in enumerate(zip(items, self.pattern.keys(), self.pattern.values())):
            if item.name == "*":
                return True
            if item.name != name:
                return False
            if item.predicate is not None and not item.predicate(value):
                return False
            if item.literal is not None and value != item.literal:
                return False
        return index + 1 == len(self.pattern)

    def into(self, pattern: str, **kwargs: str) -> Selector:
        items = _parse_follows(pattern)
        new_patterns = {}
        iterator = iter(self.pattern)
        if items and items[0].name == "~":
            if not all(item.literal or kwargs.get(item.name) for item in items[1:]):
                raise ValueError("expected specific literals in follows pattern")
            return Selector(
                {
                    **self.pattern,
                    **{item.name: kwargs.get(item.name) or item.literal for item in items[1:]},
                }  # type: ignore
            )
        for item in items:
            if item.name == "*":
                raise TypeError("expected no wildcard")
            current_key = next(iterator)
            if item.name != current_key:
                raise ValueError(f"expected {'.'.join(new_patterns)}.{item.name}, got {current_key}")
            new_patterns[current_key] = kwargs.get(item.name) or item.literal or self.pattern[current_key]
        return Selector(new_patterns)

    def expects(
        self,
        pattern: str,
        *,
        exception_type: type[Exception] = ValueError,
        **kwargs: FollowsPredicater,
    ) -> Self:
        if not self.follows(pattern, **kwargs):
            raise exception_type(f"Selector {self} does not follow {pattern}")

        return self


@runtime_checkable
class Selectable(Protocol):
    def to_selector(self) -> Selector:
        ...



from __future__ import annotations

from avilla.core.resource import LocalFileResource, RawResource
from avilla.core.ryanvk.collector.application import ApplicationCollector

from .capability import CoreCapability


class CoreResourceFetchPerform((m := ApplicationCollector())._):
    @m.entity(CoreCapability.fetch, resource=LocalFileResource)
    async def fetch_localfile(self, resource: LocalFileResource):
        return resource.file.read_bytes()

    @m.entity(CoreCapability.fetch, resource=RawResource)
    async def fetch_raw(self, resource: RawResource):
        return resource.data



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

from graia.amnesia.message import Element
from graia.amnesia.message import Text as Text

from avilla.core.resource import LocalFileResource, Resource
from avilla.core.selector import Selector


class Notice(Element):
    """该消息元素用于承载消息中用于提醒/呼唤特定用户的部分."""

    target: Selector
    display: str | None

    def __init__(self, target: Selector, display: str | None = None) -> None:
        """实例化一个 Notice 消息元素，用于承载消息中用于提醒/呼唤特定用户的部分。

        Args:
            target (str): 需要提醒/呼唤的特定用户的 ID.
            display (str, optional): 需要提醒/呼唤的特定用户的显示名称. Defaults to None.
        """

        self.target = target
        self.display = display

    def __str__(self) -> str:
        return f"[$Notice:target={self.target}]"


class NoticeAll(Element):
    """该消息元素用于群组中的管理员提醒群组中的所有成员"""

    def __init__(self) -> None:
        super().__init__()

    def __str__(self) -> str:
        return "[$NoticeAll]"


class Picture(Element):
    resource: Resource[bytes]

    def __init__(self, resource: Resource[bytes] | Path | str):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource

    def __str__(self) -> str:
        return "[$Picture]"

    def __repr__(self):
        return f"[$Picture:resource={self.resource.to_selector()}]"


class Audio(Element):
    resource: Resource[bytes]
    duration: int

    def __init__(self, resource: Resource[bytes] | Path | str, duration: int = -1):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource
        self.duration = duration

    def __str__(self) -> str:
        return "[$Audio]"

    def __repr__(self):
        return f"[$Audio:resource={self.resource.to_selector()}]"


class Video(Element):
    resource: Resource[bytes]

    def __init__(self, resource: Resource[bytes] | Path | str):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource

    def __str__(self) -> str:
        return "[$Video]"

    def __repr__(self):
        return f"[$Video:resource={self.resource.to_selector()}]"


class File(Element):
    resource: Resource[bytes]

    def __init__(self, resource: Resource[bytes] | Path | str):
        if isinstance(resource, Path):
            resource = LocalFileResource(resource)
        elif isinstance(resource, str):
            resource = LocalFileResource(Path(resource))
        self.resource = resource

    def __str__(self) -> str:
        return "[$File]"

    def __repr__(self) -> str:
        return f"[$File:resource={self.resource.to_selector()}]"


@dataclass(frozen=True, eq=True)
class Reference(Element):
    message: Selector
    slice: tuple[int, int] | None = None


@dataclass
class Face(Element):
    id: str
    name: str | None = None

    def __str__(self) -> str:
        return f"[Face:id={self.id};name={self.name}]"


class Unknown(Element):
    type: str
    raw_data: Any

    def __init__(self, type: str, raw_data: Any) -> None:
        self.type = type
        self.raw_data = raw_data

    def __str__(self) -> str:
        return f"[$Unknown:type={self.type}]"
